#! /usr/bin/env python3

import os, subprocess, shutil
from .parse import *
from .merge_control import *
from .cluster import *
from .cluster_sbnd import *
from .gather_support_read_seq import *
from .generate_consensus import *
from .locate_bp import *
from .locate_bp_sbnd import *
from .count_sread_by_alignment import *
from .post_proc import *
from .vcf_convert import *
from .insert_classify import *
from .utils import *
from .logger import get_logger
from .utils import *
from .connect import *

logger = get_logger(__name__)

def parse_main(args):

    # check if the executables exist
    is_tool("tabix")
    is_tool("bgzip")

    # check input file existences
    is_exists_bam(args.alignment_file)

    # BAM format check
    bam_cram_format_check(args.alignment_file, args.reference_fasta)

    # make directory for the output prefix
    output_dir = os.path.dirname(args.output_prefix)
    if output_dir != '' and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    ####################
    parse_alignment_info(args.alignment_file, args.reference_fasta, 
        args.output_prefix + ".tmp.deletion_info.txt", args.output_prefix + ".tmp.insertion_info.txt", 
        args.output_prefix + ".tmp.rearrangement_info.txt", args.output_prefix + ".tmp.bp_info.txt")

    ####################
    # deletion processing
    hout = open(args.output_prefix + ".tmp.deletion.sorted.bed", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k5,5n", args.output_prefix + ".tmp.deletion_info.txt"], stdout = hout)
    hout.close()

    hout = open(args.output_prefix + ".deletion.sorted.bed.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.deletion.sorted.bed"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".deletion.sorted.bed.gz"])
    ####################

    ####################
    # insertion processing
    hout = open(args.output_prefix + ".tmp.insertion.sorted.bed", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k5,5n", args.output_prefix + ".tmp.insertion_info.txt"], stdout = hout)
    hout.close()
    
    hout = open(args.output_prefix + ".insertion.sorted.bed.gz", 'w') 
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.insertion.sorted.bed"], stdout = hout)
    hout.close()
     
    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".insertion.sorted.bed.gz"])
    ####################

    ####################
    # rearrangement processing
    hout = open(args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", args.output_prefix + ".tmp.rearrangement_info.txt"], stdout = hout)
    hout.close()

    extract_bedpe_junction(args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt", 
                           args.output_prefix + ".tmp.rearrangement.bedpe",
                           args.split_alignment_check_margin, args.split_alignment_check_margin, args.minimum_breakpoint_ambiguity)

    hout = open(args.output_prefix + ".tmp.rearrangement.sorted.bedpe", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k4,4", "-k5,5n", "-k6,6n", 
                           args.output_prefix + ".tmp.rearrangement.bedpe"], stdout = hout)
    hout.close()
  
    hout = open(args.output_prefix + ".rearrangement.sorted.bedpe.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.rearrangement.sorted.bedpe"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".rearrangement.sorted.bedpe.gz"])
    ####################

    ####################
    # breakpoint processing
    hout = open(args.output_prefix + ".bp_info.sorted.bed", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", args.output_prefix + ".tmp.bp_info.txt"], stdout = hout)
    hout.close()

    hout = open(args.output_prefix + ".bp_info.sorted.bed.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".bp_info.sorted.bed"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".bp_info.sorted.bed.gz"])
    ####################

    if not args.debug:
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.deletion_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.deletion.sorted.bed"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.insertion_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.insertion.sorted.bed"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement.sorted.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.bp_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".bp_info.sorted.bed"])


def merge_control_main(args):

    merge_control_from_parse_files(args.prefix_list_file, args.output_prefix)


def call_slow_request(args, index):
    ret_code = 1
    err_message = ""
    try:
        call_slow_request_main(args, index)
        ret_code = 0
    except Exception as e:
        err_message = str(e)
    
    logger.info("End Process (%d): ret_code=%d" % (index, ret_code))
    return (ret_code, err_message)


def call_slow_request_main(args, index):

    logger.info("Generate consensus sequences (%d)" % (index))
    generate_consensus(args.tumor_prefix + ".support_read_seq.%d.txt" % (index),
        args.tumor_prefix + ".consensus_seq.%d.txt" % (index),
        use_racon = args.use_racon, debug = args.debug)
    generate_consensus_sbnd(args.tumor_prefix + ".support_read_seq.sbnd.%d.txt" % (index),
        args.tumor_prefix + ".consensus_seq.sbnd.%d.txt" % (index),
        use_racon = args.use_racon, debug = args.debug, max_memory_minimap2 = args.max_memory_minimap2)

    logger.info("Locating single-base resolution break points for candidate SVs (%d)" % (index))
    locate_bp(args.tumor_prefix + ".consensus_seq.%d.txt" % (index),
        args.tumor_prefix + ".refined_bp.%d.txt" % (index),
        args.reference_fasta, args.sw_jump_params, args.debug)
    locate_bp_sbnd(args.tumor_prefix + ".consensus_seq.sbnd.%d.txt" % (index),
        args.tumor_prefix + ".refined_bp.sbnd.%d.txt" % (index),
        args.reference_fasta, args.debug)

    logger.info("Counting the number of supporting read for the tumor by realignment of SV candidate segments (%d)" % (index))
    count_sread_by_alignment(
        args.tumor_prefix + ".refined_bp.%d.txt" % (index), 
        args.tumor_bam, 
        args.tumor_prefix + ".realignment.tumor.sread_count.%d.txt" % (index), 
        args.tumor_prefix + ".realignment.tumor.sread_info.%d.txt" % (index), 
        args.reference_fasta,
        sbnd_file = args.tumor_prefix + ".refined_bp.sbnd.%d.txt" % (index), 
        output_count_file_sbnd = args.tumor_prefix + ".realignment.tumor.sread_count.sbnd.%d.txt" % (index),
        output_alignment_info_file_sbnd = args.tumor_prefix + ".realignment.tumor.sread_info.sbnd.%d.txt" % (index),
        check_read_max_num = args.check_read_max_num, 
        var_read_min_mapq = args.var_read_min_mapq, score_ratio_thres = args.validation_score_ratio_thres, use_ssw_lib = False, 
        sort_option = args.sort_option, debug = args.debug
    )
     
    if args.control_bam is not None:
        logger.info("Counting the number of supporting read for the control by realignment of SV candidate segments (%d)" % (index))
        count_sread_by_alignment(
            args.tumor_prefix + ".refined_bp.%d.txt" % (index), 
            args.control_bam, 
            args.tumor_prefix + ".realignment.control.sread_count.%d.txt" % (index), 
            args.tumor_prefix + ".realignment.control.sread_info.%d.txt" % (index), 
            args.reference_fasta,
            sbnd_file = args.tumor_prefix + ".refined_bp.sbnd.%d.txt" % (index), 
            output_count_file_sbnd = args.tumor_prefix + ".realignment.control.sread_count.sbnd.%d.txt" % (index),
            output_alignment_info_file_sbnd = args.tumor_prefix + ".realignment.control.sread_info.sbnd.%d.txt" % (index),
            check_read_max_num = args.check_read_max_num, 
            var_read_min_mapq = args.var_read_min_mapq, score_ratio_thres = args.validation_score_ratio_thres, use_ssw_lib = False, 
            sort_option = args.sort_option, debug = args.debug
        )


def get_main(args):

    # check if the executables exist
    if args.use_racon: 
        is_tool("racon")
    else:
        is_tool("mafft")
    if False: libssw_check()

    if args.single_bnd: 
        is_tool("minimap2")
        if not args.use_racon:
            logger.error("single_bnd option has to be used with use_racon option")
            sys.exit(1)

    parallel_num = 1
    # if args.threads > 1:
    #     if args.processes > 1:
    #         logger.error("--threads option sholud not be used with processes option")
    #         sys.exit(1)
    #     logger.warning("--threads option is not recommended. Please consider using --processes option")
    #     parallel_num = args.threads
    # elif args.processes > 1:
    if args.processes > 1:
        parallel_num = args.processes

    # parameter preset 
    if sum([int(x == True) for x in [args.qv10, args.qv15, args.qv20, args.qv25]]) > 1:
        logger.error("Parameter preset (qv10, qv15, qv20, qv25) should be set only once")
        sys.exit(1)

    if args.qv10:
        args.validation_score_ratio_thres = 1.2
        args.sw_jump_params = [1, 2, 2, 2]
    elif args.qv15:
        args.validation_score_ratio_thres = 1.4
        args.sw_jump_params = [1, 3, 3, 2]
    elif args.qv20: 
        args.validation_score_ratio_thres = 1.6
        args.sw_jump_params = [1, 4, 4, 2]
    elif args.qv25:
        args.validation_score_ratio_thres = 1.8
        args.sw_jump_params = [1, 6, 6, 2]

    # check existences
    is_exists_bam(args.tumor_bam)
    is_exists(args.reference_fasta)
    if args.control_bam is not None: is_exists_bam(args.control_bam)
    if args.simple_repeat_bed is not None: is_exists(args.simple_repeat_bed)
   
    # check parsed files existences
    is_exists_parsed_files(args.tumor_prefix)
    if args.control_prefix is not None: is_exists_parsed_files(args.control_prefix)
    if args.control_panel_prefix is not None: is_exists_parsed_files(args.control_panel_prefix)
 
    # BAM format check
    bam_cram_format_check(args.tumor_bam, args.reference_fasta)
    fasta_format_check(args.reference_fasta)
    if args.control_bam is not None: bam_cram_format_check(args.control_bam, args.reference_fasta)

    control_rearrangement_bedpe = None
    control_deletion_bed = None
    control_insertion_bed = None
    control_bp_bed = None
    if args.control_prefix is not None: 
        control_rearrangement_bedpe = args.control_prefix + ".rearrangement.sorted.bedpe.gz"
        control_deletion_bed = args.control_prefix + ".deletion.sorted.bed.gz"
        control_insertion_bed = args.control_prefix + ".insertion.sorted.bed.gz"
        # if args.single_bnd is not None:
        control_bp_bed = args.control_prefix + ".bp_info.sorted.bed.gz"


    control_panel_rearrangement_bedpe = None
    control_panel_deletion_bed = None
    control_panel_insertion_bed = None
    control_panel_bp_bed = None
    if args.control_panel_prefix is not None:
        control_panel_rearrangement_bedpe = args.control_panel_prefix + ".rearrangement.sorted.bedpe.gz"
        control_panel_deletion_bed = args.control_panel_prefix + ".deletion.sorted.bed.gz"
        control_panel_insertion_bed = args.control_panel_prefix + ".insertion.sorted.bed.gz"
        # if args.single_bnd is not None:
        control_panel_bp_bed = args.control_panel_prefix + ".bp_info.sorted.bed.gz"

    bp_bed = None
    if args.use_racon:
        bp_bed = args.tumor_prefix + ".bp_info.sorted.bed.gz"
    ####################
    logger.info("Clustering rearrangement type supporting reads for putative SVs")
    cluster_supporting_reads(args.tumor_prefix + ".rearrangement.sorted.bedpe.gz", 
        args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe",
        "rearrangement", control_junction_bedpe = control_rearrangement_bedpe,
        control_panel_junction_bedpe = control_panel_rearrangement_bedpe,
        read_num_thres = args.min_tumor_variant_read_num, cluster_margin_size = args.cluster_margin_size, 
        median_mapQ_thres = args.median_mapQ_thres, max_overhang_size_thres = args.max_overhang_size_thres,
        maximum_local_variant_num = 1000, max_control_read_num = args.max_control_variant_read_num, 
        max_panel_read_num = args.max_panel_read_num, max_panel_sample_num = args.max_panel_sample_num, debug = args.debug)

    logger.info("Clustering insertion type supporting reads for putative SVs")
    cluster_supporting_reads(args.tumor_prefix + ".insertion.sorted.bed.gz",
        args.tumor_prefix + ".insertion.sorted.clustered.bedpe",
        "insertion", control_junction_bedpe = control_insertion_bed, 
        control_panel_junction_bedpe = control_panel_insertion_bed, bp_bed = bp_bed,
        read_num_thres = args.min_tumor_variant_read_num, cluster_margin_size = args.cluster_margin_size, 
        median_mapQ_thres = args.median_mapQ_thres, max_overhang_size_thres = args.max_overhang_size_thres,
        min_indel_size = int(0.7 * args.min_indel_size), max_control_read_num = args.max_control_variant_read_num, 
        max_panel_read_num = args.max_panel_read_num, max_panel_sample_num = args.max_panel_sample_num, debug = args.debug)

    logger.info("Clustering deletion type supporting reads for putative SVs")
    cluster_supporting_reads(args.tumor_prefix + ".deletion.sorted.bed.gz",
        args.tumor_prefix + ".deletion.sorted.clustered.bedpe",
        "deletion", control_junction_bedpe = control_deletion_bed, 
        control_panel_junction_bedpe = control_panel_deletion_bed, bp_bed = bp_bed,
        read_num_thres = args.min_tumor_variant_read_num, cluster_margin_size = args.cluster_margin_size, 
        median_mapQ_thres = args.median_mapQ_thres, max_overhang_size_thres = args.max_overhang_size_thres,
        min_indel_size = int(0.7 * args.min_indel_size), max_control_read_num = args.max_control_variant_read_num, 
        max_panel_read_num = args.max_panel_read_num, max_panel_sample_num = args.max_panel_sample_num, debug = args.debug)

    if args.single_bnd:
        logger.info("Clustering single breakend type supporting reads for putative SVs")
        cluster_supporting_reads_sbnd(args.tumor_prefix + ".bp_info.sorted.bed.gz",
            args.tumor_prefix + ".singlebreakend.sorted.clustered.bed", control_bed = control_bp_bed,
            control_panel_bed = control_panel_bp_bed, 
            read_num_thres = args.min_tumor_variant_read_num, cluster_margin_size = args.cluster_margin_size,
            median_mapQ_thres = args.median_mapQ_thres, max_control_read_num = args.max_control_variant_read_num,
            max_panel_read_num = args.max_panel_read_num, max_panel_sample_num = args.max_panel_sample_num, debug = args.debug)
    else:
        with open(args.tumor_prefix + ".singlebreakend.sorted.clustered.bed", 'w'):
            pass

    logger.info("Gathering sequences of supporting reads")
    gather_support_read_seq(args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe",
        args.tumor_prefix + ".insertion.sorted.clustered.bedpe",
        args.tumor_prefix + ".deletion.sorted.clustered.bedpe",
        args.tumor_prefix + ".support_read_seq.txt",
        args.tumor_bam, args.reference_fasta, single_breakend_file = args.tumor_prefix + ".singlebreakend.sorted.clustered.bed",
        output_file_sbind = args.tumor_prefix + ".support_read_seq.sbnd.txt" )
    
    logger.info("Preparation for parallel execution")
    fw_support_read_seqs = []
    fw_support_read_seq_sbnds = []
    for i in range(parallel_num):
        fw_support_read_seqs.append(open(args.tumor_prefix + ".support_read_seq.%d.txt" % (i), "w"))
        fw_support_read_seq_sbnds.append(open(args.tumor_prefix + ".support_read_seq.sbnd.%d.txt" % (i), "w"))

    last_key = ""
    last_tpos = -1
    with open(args.tumor_prefix + ".support_read_seq.txt") as hin:
        for row in hin:
            if row.rstrip() == "": continue
            key = row.split("\t")[0]
            if last_key != key:
                last_key = key
                last_tpos += 1
                if last_tpos == parallel_num:
                    last_tpos = 0
            fw_support_read_seqs[last_tpos].write(row)
    
    last_key = ""
    last_tpos = -1
    with open(args.tumor_prefix + ".support_read_seq.sbnd.txt") as hin:
        for row in hin:
            if row.rstrip() == "": continue
            key = row.split("\t")[0]
            if last_key != key:
                last_key = key
                last_tpos += 1
                if last_tpos == parallel_num:
                    last_tpos = 0
            fw_support_read_seq_sbnds[last_tpos].write(row)

    for i in range(parallel_num):
        fw_support_read_seqs[i].close()
        fw_support_read_seq_sbnds[i].close()

    if parallel_num == 1:
        call_slow_request_main(args, 0)
    else:
        import concurrent.futures

        if args.processes > 1:
            with concurrent.futures.ProcessPoolExecutor(max_workers=parallel_num) as executor:
                features = [executor.submit(call_slow_request, args, i) for i in range(parallel_num)]
                for feature in features:
                    (ret_code, err_message) = feature.result()
                    if ret_code != 0:
                        raise Exception(err_message)
        else:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                features = [executor.submit(call_slow_request, args, i) for i in range(parallel_num)]
                for feature in features:
                    (ret_code, err_message) = feature.result()
                    if ret_code != 0:
                        raise Exception(err_message)

    logger.info("Merge parallel execution")

    """
    def merge_txt(prefix):
        with open(prefix + ".txt", 'w') as hout:
            for i in range(parallel_num):
                for l in open(prefix + ".%d.txt" % (i)):
                    hout.write(l)
    """

    def merge_txt(prefix):
        with open(prefix + ".txt", 'w') as hout:
            for i in range(parallel_num):
                with open(prefix + ".%d.txt" % (i), 'r') as hin:
                    for l in hin:
                        hout.write(l)

    merge_txt(args.tumor_prefix + ".refined_bp.sbnd")
    merge_txt(args.tumor_prefix + ".realignment.tumor.sread_count")
    merge_txt(args.tumor_prefix + ".realignment.tumor.sread_info")
    merge_txt(args.tumor_prefix + ".realignment.tumor.sread_count.sbnd")
    merge_txt(args.tumor_prefix + ".realignment.tumor.sread_info.sbnd")
    if args.control_bam is not None:
        merge_txt(args.tumor_prefix + ".realignment.control.sread_count")
        merge_txt(args.tumor_prefix + ".realignment.control.sread_info")
        merge_txt(args.tumor_prefix + ".realignment.control.sread_count.sbnd")
        merge_txt(args.tumor_prefix + ".realignment.control.sread_info.sbnd")
    if not args.debug:
        for i in range(parallel_num):
            os.remove(args.tumor_prefix + ".support_read_seq.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".support_read_seq.sbnd.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".consensus_seq.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".consensus_seq.sbnd.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".refined_bp.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".refined_bp.sbnd.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".realignment.tumor.sread_count.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".realignment.tumor.sread_count.sbnd.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".realignment.tumor.sread_info.%d.txt" % (i))
            os.remove(args.tumor_prefix + ".realignment.tumor.sread_info.sbnd.%d.txt" % (i))
            if args.control_bam is not None:
                os.remove(args.tumor_prefix + ".realignment.control.sread_count.%d.txt" % (i))
                os.remove(args.tumor_prefix + ".realignment.control.sread_count.sbnd.%d.txt" % (i))
                os.remove(args.tumor_prefix + ".realignment.control.sread_info.%d.txt" % (i))
                os.remove(args.tumor_prefix + ".realignment.control.sread_info.sbnd.%d.txt" % (i))

    logger.info("Final processing") 
    control_sread_count_file = args.tumor_prefix + ".realignment.control.sread_count.txt" if args.control_bam is not None else None
    integrate_realignment_result(args.tumor_prefix + ".realignment.tumor.sread_count.txt", control_sread_count_file,
        args.tumor_prefix + ".nanomonsv.result.txt", args.reference_fasta, args.simple_repeat_bed, min_indel_size = args.min_indel_size,
        min_tumor_variant_read_num = args.min_tumor_variant_read_num, min_tumor_VAF = args.min_tumor_VAF, 
        max_control_variant_read_num = args.max_control_variant_read_num, max_control_VAF = args.max_control_VAF)

    genomesv2vcf_convert(args.tumor_prefix + ".nanomonsv.result.txt", args.tumor_prefix + ".nanomonsv.result.vcf", 
        args.reference_fasta)

    proc_sread_info_file(args.tumor_prefix + ".realignment.tumor.sread_info.txt",
        args.tumor_prefix + ".nanomonsv.result.txt",
        args.tumor_prefix + ".nanomonsv.supporting_read.txt")

    control_sread_count_file_sbnd = args.tumor_prefix + ".realignment.control.sread_count.sbnd.txt" if args.control_bam is not None else None
    integrate_realignment_result_sbnd(args.tumor_prefix + ".realignment.tumor.sread_count.sbnd.txt", control_sread_count_file_sbnd,
        args.tumor_prefix + ".nanomonsv.sbnd.result.txt",
        args.tumor_prefix + ".nanomonsv.result.txt", args.tumor_prefix + ".refined_bp.sbnd.txt")
 
    if not args.debug:
        os.remove(args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe")
        os.remove(args.tumor_prefix + ".insertion.sorted.clustered.bedpe")
        os.remove(args.tumor_prefix + ".deletion.sorted.clustered.bedpe")
        os.remove(args.tumor_prefix + ".singlebreakend.sorted.clustered.bed")
        os.remove(args.tumor_prefix + ".support_read_seq.txt")
        os.remove(args.tumor_prefix + ".support_read_seq.sbnd.txt")
        #os.remove(args.tumor_prefix + ".consensus_seq.txt")
        #os.remove(args.tumor_prefix + ".consensus_seq.sbnd.txt")
        #os.remove(args.tumor_prefix + ".refined_bp.txt")
        os.remove(args.tumor_prefix + ".refined_bp.sbnd.txt")
        os.remove(args.tumor_prefix + ".realignment.tumor.sread_count.txt")
        os.remove(args.tumor_prefix + ".realignment.tumor.sread_count.sbnd.txt")
        os.remove(args.tumor_prefix + ".realignment.tumor.sread_info.txt")  
        os.remove(args.tumor_prefix + ".realignment.tumor.sread_info.sbnd.txt")
        if args.control_bam is not None:
            os.remove(args.tumor_prefix + ".realignment.control.sread_count.txt")
            os.remove(args.tumor_prefix + ".realignment.control.sread_count.sbnd.txt")
            os.remove(args.tumor_prefix + ".realignment.control.sread_info.txt")
            os.remove(args.tumor_prefix + ".realignment.control.sread_info.sbnd.txt")

        if not args.single_bnd:
            os.remove(args.tumor_prefix + ".nanomonsv.sbnd.result.txt")   


def validate_main(args):
   
    # executable check
    # if False: libssw_check()

    # parameter preset 
    if sum([int(x == True) for x in [args.qv10, args.qv15, args.qv20, args.qv25]]) > 1:
        logger.error("Parameter preset (qv10, qv15, qv20, qv25) should be set only once")
        sys.exit(1)

    if args.qv10:
        args.validation_score_ratio_thres = 1.2
    elif args.qv15:
        args.validation_score_ratio_thres = 1.4
    elif args.qv20: 
        args.validation_score_ratio_thres = 1.6
    elif args.qv25:
        args.validation_score_ratio_thres = 1.8

    # check existences
    is_exists_bam(args.tumor_bam)
    is_exists(args.reference_fasta)
    if args.control_bam is not None: is_exists_bam(args.control_bam)

    # BAM format check
    bam_cram_format_check(args.tumor_bam, args.reference_fasta)
    fasta_format_check(args.reference_fasta)
    if args.control_bam is not None: bam_cram_format_check(args.control_bam, args.reference_fasta)

    
    logger.info("Counting the number of supporting read for the tumor by realignment of SV candidate segments")
    count_sread_by_alignment(args.sv_list_file, args.tumor_bam,
        args.output + ".realignment.tumor.sread_count.txt", args.output + ".realignment.tumor.sread_info.txt",
        args.reference_fasta, var_read_min_mapq = args.var_read_min_mapq, use_ssw_lib = False, 
        sort_option = args.sort_option, debug = args.debug)

    if args.control_bam is not None:
        logger.info("Counting the number of supporting read for the control by realignment of SV candidate segments")
        count_sread_by_alignment(args.sv_list_file, args.control_bam,
            args.output + ".realignment.control.sread_count.txt", args.output + ".realignment.control.sread_info.txt",
            args.reference_fasta, var_read_min_mapq = args.var_read_min_mapq, use_ssw_lib = False, 
            sort_option = args.sort_option, debug = args.debug)

    logger.info("Final processing")
    control_sread_count_file = args.output + ".realignment.control.sread_count.txt" if args.control_bam is not None else None
    integrate_realignment_result(args.output + ".realignment.tumor.sread_count.txt", control_sread_count_file, args.output,
        args.reference_fasta, min_indel_size = 0, min_tumor_variant_read_num = 0, min_tumor_VAF = 0,
        max_control_variant_read_num = float("inf"), max_control_VAF = float("inf"))

    if not args.debug:
        os.remove(args.output + ".realignment.tumor.sread_count.txt")
        os.remove(args.output + ".realignment.tumor.sread_info.txt")
        os.remove(args.output + ".realignment.control.sread_count.txt")
        os.remove(args.output + ".realignment.control.sread_info.txt")
    ####################
    

def insert_classify_main(args):

    import tempfile
    # import annot_utils.exon

    # check if the executables exist
    is_tool("minimap2")
    is_tool("bedtools")
    is_tool("bwa")
    is_tool("RepeatMasker")

    make_fasta_file(args.sv_list_file, args.output_file + ".tmp.fasta", args.output_file + ".tmp.seq_id.txt")
   
    ##########
    # processed pseudo gene
    # annot_utils.exon.make_exon_info(args.output_file + ".tmp.exon.bed.gz", "gencode", args.genome_id, args.grc, True)

    make_exon_bed_from_gtf(args.gtf_file, args.output_file + ".tmp.exon.bed.gz")

    with open(args.output_file + ".tmp.minimap2.sam", 'w') as hout:
        subprocess.check_call(["minimap2", "-ax", "splice", args.reference_fasta, args.output_file + ".tmp.fasta"], stdout = hout)

    sam2bed_split(args.output_file + ".tmp.minimap2.sam", args.output_file + ".tmp.minimap2.filt.bed")

    with open(args.output_file + ".tmp.minimap2.filt.exon.bed", 'w') as hout:
        subprocess.check_call(["bedtools", "intersect", "-a", args.output_file + ".tmp.minimap2.filt.bed", 
                               "-b", args.output_file + ".tmp.exon.bed.gz", "-wo"], stdout = hout)
 
    pp_proc_filt_exon(args.output_file + ".tmp.minimap2.filt.exon.bed", 
                      args.output_file + ".tmp.seq_id.txt", 
                      args.output_file + ".tmp.ppseudo.txt")
    ##########

    ##########
    # repeat masker
    output_dir = os.path.dirname(args.output_file)
    tmpdir_rmsk = tempfile.mkdtemp()
    # tmpdir_rmsk = "tmp"
    subprocess.check_call(["RepeatMasker", "-species", "human", args.output_file + ".tmp.fasta", "-dir", tmpdir_rmsk])

    summarize_rmsk(tmpdir_rmsk + '/' + os.path.basename(args.output_file + ".tmp.fasta") + ".out", args.output_file + ".tmp.rmsk.txt")

    check_tsd_polyAT(args.output_file + ".tmp.fasta", args.output_file + ".tmp.seq_id.txt", 
                     args.reference_fasta, args.output_file + ".tmp.tsd.polyAT.txt")

    shutil.rmtree(tmpdir_rmsk)
    ##########

    ##########
    # alignment to reference genome
    with open(args.output_file + ".tmp.bwa.sam", 'w') as hout:
        print(' '.join(["bwa", "mem", "-h", "200", args.reference_fasta, args.output_file + ".tmp.fasta"]))
        subprocess.check_call(["bwa", "mem", "-h", "200", args.reference_fasta, args.output_file + ".tmp.fasta"], stdout = hout)

    summarize_bwa_alignment2(args.output_file + ".tmp.bwa.sam", args.output_file + ".tmp.seq_id.txt", args.output_file + ".tmp.alignment.txt")

    organize_info(args.output_file + ".tmp.rmsk.txt", args.output_file + ".tmp.alignment.txt", 
                 args.output_file + ".tmp.tsd.polyAT.txt", args.output_file + ".tmp.seq_id.txt", 
                 args.output_file + ".tmp.org.txt", args.LINE1_db)

    annotate_sv_file(args.sv_list_file, args.output_file + ".tmp.org.txt", args.output_file + ".tmp.ppseudo.txt",
                     args.output_file + ".tmp.seq_id.txt", args.output_file)

    if not args.debug:
        os.remove(args.output_file + ".tmp.fasta")
        os.remove(args.output_file + ".tmp.seq_id.txt")
        os.remove(args.output_file + ".tmp.bwa.sam")
        os.remove(args.output_file + ".tmp.exon.bed.gz")
        os.remove(args.output_file + ".tmp.exon.bed.gz.tbi")
        os.remove(args.output_file + ".tmp.minimap2.filt.bed")
        os.remove(args.output_file + ".tmp.minimap2.filt.exon.bed")
        os.remove(args.output_file + ".tmp.ppseudo.txt")
        os.remove(args.output_file + ".tmp.rmsk.txt")
        os.remove(args.output_file + ".tmp.tsd.polyAT.txt")
        os.remove(args.output_file + ".tmp.minimap2.sam")
        os.remove(args.output_file + ".tmp.alignment.txt")
        os.remove(args.output_file + ".tmp.org.txt")

def connect_main(args):
    hout = open(args.output_prefix + ".supporting_read.sorted.txt", 'w')
    subprocess.check_call(["sort", "-k9,9", "-k10,10n", args.support_read_file], stdout = hout)
    hout.close()
    filter_support_reads(args.nanomonsv_result_file, args.output_prefix + ".supporting_read.sorted.txt", args.output_prefix + ".supporting_read.sorted.filtered.txt")
    get_edge(args.output_prefix + ".supporting_read.sorted.filtered.txt", args.output_prefix + ".edge.txt", args.consistent_pos_margin)
    connect(args.output_prefix + ".edge.txt", args.output_prefix + ".connect.txt")
    
    if not args.debug:
        os.remove(args.output_prefix + ".supporting_read.sorted.txt")



#! /usr/bin/env python

import sys, gzip
import pysam

rmsk_file = sys.argv[1]

with gzip.open(rmsk_file, 'rt') as hin:
    for line in hin:
        F = line.rstrip('\n').split('\t')
        # if len(F[5]) > 5: continue
        if int(F[7]) - int(F[6]) < 5800: continue
        if F[12] != "L1": continue
        if not F[10] in ["L1HS", "L1PA2", "L1PA3", "L1PA4", "L1PA5"]: continue
        label = ','.join([F[5], F[6], F[7], F[9], F[10]])
        print('\t'.join([F[5], F[6], F[7], label, '0', F[9]]))
 



#! /usr/bin/env pytho3

import sys, subprocess, os, shutil, csv
from collections import namedtuple
import pysam
from nanomonsv.post_proc import *

"""
def make_fasta(input_file, output_prefix):

    with open(input_file, 'r') as hin, open(output_prefix, 'w') as hout:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            print(f">{F[0]},{F[1]},{F[2]},{F[4]},{len(F[3])}\n{F[3]}", file = hout)        
"""

def proc_rmsk(input_file, output_file):

    with open(input_file, 'r') as hin, open(output_file, 'w') as hout:
        print("Contig_ID\tContig_Len\tQuery_Align_Start\tQuery_Align_End\tQuery_Align_Strand\tRepeat_NAME\tRepeat_Class\tRepeat_Align_Start\tRepeat_Align_End", file = hout)
        for line in hin:
            F = line.rstrip('\n').split()
            if len(F) <= 1 or F[0] in ["SW", "score"]: continue
            if F[0] == "There": continue

            contigs = F[4].split(',')
            contig_id, contig_len = ','.join(contigs[:-1]), contigs[-1]
            if F[8] == 'C': F[8] = '-'
            if F[8] == '+':
                print(f"{contig_id}\t{contig_len}\t{F[5]}\t{F[6]}\t{F[8]}\t{F[9]}\t{F[10]}\t{F[11]}\t{F[12]}", file = hout)
            else:
                print(f"{contig_id}\t{contig_len}\t{F[5]}\t{F[6]}\t{F[8]}\t{F[9]}\t{F[10]}\t{F[13]}\t{F[12]}", file = hout)


def proc_sam(input_sam, output_file):

    samfile = pysam.AlignmentFile(input_sam, 'r')

    query2align_supp = {}
    query2align_primary = {}
    query2align_ins = {}

    hout = open(output_file, 'w') 
    print("Contig_ID\tContig_Len\tQuery_Align_Start\tQuery_Align_End\tQuery_Align_Strand\tTarget_Align_Chromosome\tTarget_Align_Start\tTarget_Align_End\tMapping_Quality", file = hout)
    for read in samfile.fetch():

        if read.is_unmapped: continue
        if read.is_secondary: continue

        query_name = read.query_name
        key = query_name

        query_strand = '-' if read.is_reverse else '+'
        reference_name = read.reference_name
        reference_start = read.reference_start + 1
        reference_end = read.reference_end
        mapping_quality = read.mapping_quality
        query_length = read.infer_read_length()

        cigartuples = read.cigartuples
        left_hard_clipping_size, right_hard_clipping_size = 0, 0
        total_ins_size, total_del_size = 0, 0
        if cigartuples[0][0] == 5: left_hard_clipping_size = cigartuples[0][1]
        if cigartuples[-1][0] == 5: right_hard_clipping_size = cigartuples[-1][1]

        if not read.is_supplementary:
            if query_strand == '+':
                query_start = read.query_alignment_start + 1 
                query_end = read.query_alignment_end
            else:
                query_start = query_length - read.query_alignment_end + 1
                query_end = query_length - read.query_alignment_start
        else:
            if query_strand == '+':
                query_start = left_hard_clipping_size + 1
                query_end = query_length - right_hard_clipping_size
            else:
                query_start = right_hard_clipping_size + 1
                query_end = query_length - left_hard_clipping_size

        contigs = query_name.split(',')
        contig_id, contig_len = ','.join(contigs[:-1]), contigs[-1]
        print(f"{contig_id}\t{contig_len}\t{query_start}\t{query_end}\t{query_strand}\t{reference_name}\t{reference_start}\t{reference_end}\t{mapping_quality}", file = hout)

    hout.close()



Rmsk_record = namedtuple("Rmsk_record", "QA_Start QA_End QA_Strand RName RClass RA_Start RA_End")

Bwa_record = namedtuple("Bwa_record", "QA_Start QA_End QA_Strand TA_Chr TA_Start TA_End MQ")

def reverse_complement(seq):
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
                  'W': 'W', 'S': 'S', 'M': 'K', 'K': 'M', 'R': 'Y', 'Y': 'R',
                  'B': 'V', 'V': 'B', 'D': 'H', 'H': 'D', 'N': 'N'}

    return("".join(complement.get(base, base) for base in reversed(seq)))

class Contig_info:
    
    def __init__(self, contig_id, contig):

        bchr, bpos, bstrand, bid = contig_id.split(',')
        
        self.bp_chr = bchr
        self.bp_pos = int(bpos)
        self.bp_strand = bstrand
        self.bp_id = bid
        self.contig = contig
        self.contig_len = len(contig)
        self.contig_class = "None"
        self.rmsk_info = []
        self.bwa_info = []        

        self.chr1 = bchr 
        self.chr2 = None
        self.pos1 = int(bpos)
        self.pos2 = None
        self.dir1 = bstrand
        self.dir2 = None
        self.inseq = None

        self.repeat_size_ratio_thres = 0.8

    def sv_key(self):

        if self.chr2 is not None:

            if self.inseq == '': self.inseq = '---'

            if self.chr1 > self.chr2 or (self.chr1 == self.chr2 and self.pos1 > self.pos2):
                self.chr1, self.chr2 = self.chr2, self.chr1
                self.pos1, self.pos2 = self.pos2, self.pos1
                self.dir1, self.dir2 = self.dir2, self.dir1
                self.inseq = reverse_complement(self.inseq)

            return(f'{self.chr1},{self.pos1},{self.dir1},{self.chr2},{self.pos2},{self.dir2},{self.inseq}')
        else:
            return("---")


    def check_classification(self):

        self.check_simple_satellite()
        if self.contig_class == "None":
            self.check_plain_SV()
        if self.contig_class == "None":
            self.check_L1_mediated_deletion()
        if self.contig_class == "None":
            self.contig_class = "Complex"

    def check_simple_satellite(self):

        simple_repeat_size = 0
        satellite_size = 0
        satellite_alpha_size = 0
        for rec in self.rmsk_info:
            if rec.RClass == "Simple_repeat":
                simple_repeat_size = simple_repeat_size + rec.QA_End - rec.QA_Start
            if rec.RClass == "Satellite":
                satellite_size = satellite_size + rec.QA_End - rec.QA_Start
            if rec.RClass == "Satellite/centr":
                satellite_alpha_size = satellite_alpha_size + rec.QA_End - rec.QA_Start

        if simple_repeat_size / self.contig_len > self.repeat_size_ratio_thres:
            self.contig_class = "Simple_repeat"
        if satellite_size / self.contig_len > self.repeat_size_ratio_thres:
            self.contig_class = "Satellite"
        if satellite_alpha_size / self.contig_len > self.repeat_size_ratio_thres:
            self.contig_class = "Satellite/centr"


    def check_plain_SV(self):

        for rec in self.bwa_info:
            if rec.QA_Start < 100 and rec.QA_End - rec.QA_Start >= 2000 and rec.MQ >= 40:
                self.contig_class = "Plain_SV"

                self.chr2 = rec.TA_Chr
                if rec.QA_Strand == '+':
                    self.pos2 = rec.TA_Start
                    self.dir2 = '-'
                else:
                    self.pos2 = rec.TA_End
                    self.dir2 = '+'

                self.inseq = self.contig[:(rec.QA_Start - 1)] 

    def check_L1_mediated_deletion(self):

        self.bwa_info = sorted(self.bwa_info, key = lambda x: x.QA_Start)
        
        L1HS_segments = []
        for rec in self.rmsk_info:
            if rec.RName in ["L1HS", "L1P1", "L1PA2"]: L1HS_segments.append((rec.QA_Start, rec.QA_End))

        L1HS_intersection_size = 0
        query_size = 0
        for i in range(0, min(len(self.bwa_info) - 1, 2)):
            rec = self.bwa_info[i]
            query_size = query_size + rec.QA_End - rec.QA_Start
            for lseg in L1HS_segments:
                if lseg[1] >= rec.QA_Start and lseg[0] <= rec.QA_End:
                    L1HS_intersection_size = L1HS_intersection_size + min(lseg[1], rec.QA_End) - max(lseg[0], rec.QA_Start)
    
            if float(L1HS_intersection_size) / query_size > 0.8:
                if self.bwa_info[i + 1].QA_End - self.bwa_info[i + 1].QA_Start >= 2000 and self.bwa_info[i + 1].MQ >= 40:
                    self.contig_class = "L1_Mediated_Del" 

                    self.chr2 = self.bwa_info[i + 1].TA_Chr
                    if self.bwa_info[i + 1].QA_Strand == '+':
                        self.pos2 = self.bwa_info[i + 1].TA_Start
                        self.dir2 = '-'
                    else:
                        self.pos2 = self.bwa_info[i + 1].TA_End
                        self.dir2 = '+'
                        
                    self.inseq = self.contig[:(self.bwa_info[i + 1].QA_Start - 1)]

                elif self.bwa_info[i + 1].QA_End - self.bwa_info[i + 1].QA_Start >= 2000 and self.bwa_info[i + 1].TA_Chr == self.bp_chr:
                    if (self.bwa_info[i + 1].QA_Strand == self.bp_strand == '+' and -100 < self.bwa_info[i + 1].TA_Start - self.bp_pos < 10000) or \
                        (self.bwa_info[i + 1].QA_Strand == self.bp_strand == '-' and -10000 < self.bwa_info[i + 1].TA_End - self.bp_pos < 100):
                        self.contig_class = "L1_Mediated_Del"

                        self.chr2 = self.bwa_info[i + 1].TA_Chr
                        if self.bwa_info[i + 1].QA_Strand == '+':
                            self.pos2 = self.bwa_info[i + 1].TA_Start
                            self.dir2 = '-'
                        else:
                            self.pos2 = self.bwa_info[i + 1].TA_End
                            self.dir2 = '+'

                        self.inseq = self.contig[:(self.bwa_info[i + 1].QA_Start - 1)]


def key_check(tchr, tpos, tdir, key2info, non_string, margin = 50):

    for key in key2info:

        if tchr == key[0] and tdir == key[2]:
            if abs(int(tpos) - int(key[1])) < margin:

                return key2info[key]

    return non_string



if __name__ == "__main__":
   
 
    # normal_sv_input_file = sys.argv[1] 
    # sbnd_sv_input_file = sys.argv[2]
    output_prefix = sys.argv[1]
    reference_fasta = sys.argv[2]

    nanomonsv2info = {}
    with open(output_prefix + ".nanomonsv.result.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = '\t'):
            tinfo = f'TRUE\t{F["Chr_1"]},{F["Pos_2"]},{F["Dir_1"]},{F["Chr_2"]},{F["Pos_2"]},{F["Dir_2"]}'
            key = (F["Chr_1"], F["Pos_1"], F["Dir_1"])
            nanomonsv2info[key] = tinfo
            key = (F["Chr_2"], F["Pos_2"], F["Dir_2"])       
            nanomonsv2info[key] = tinfo

    with open(output_prefix + ".nanomonsv.sbnd.result.txt", 'r') as hin, open(output_prefix + ".tmp.fasta", 'w') as hout:
        for F in csv.DictReader(hin, delimiter = "\t"):
            nanomonsv_str = key_check(F["Chr_1"], F["Pos_1"], F["Dir_1"], nanomonsv2info, "FALSE")
            if nanomonsv_str.startswith("FALSE"): 
                print(f'>{F["Chr_1"]},{F["Pos_1"]},{F["Dir_1"]},{F["SV_ID"]},{len(F["Contig"])}\n{F["Contig"]}', file = hout)

    os.makedirs(output_prefix + ".tmp.rmsk", exist_ok = True)
    subprocess.check_call(["RepeatMasker", "-species", "human", output_prefix + ".tmp.fasta", "-dir", output_prefix + ".tmp.rmsk"])

    boutput_prefix = os.path.basename(output_prefix)
    proc_rmsk(output_prefix + ".tmp.rmsk/" + boutput_prefix + ".tmp.fasta" + ".out", output_prefix + ".nanomonsv.rmsk.txt")

    with open(output_prefix + ".tmp.bwa.sam", 'w') as hout:
        subprocess.check_call(["bwa", "mem", "-h", "200", reference_fasta, output_prefix + ".tmp.fasta"], stdout = hout)
   
    proc_sam(output_prefix + ".tmp.bwa.sam", output_prefix + ".nanomonsv.bwa.txt")

    shutil.rmtree(output_prefix + ".tmp.rmsk")
    os.remove(output_prefix + ".tmp.fasta")
    os.remove(output_prefix + ".tmp.bwa.sam")

    contig_id2contig = {}                    
    contig_id2contig_info = {}
    
    contig2L1HS_info = {}
    simple_repeat_size = 0
    satellite_size = 0

    with open(output_prefix + ".nanomonsv.sbnd.result.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = '\t'):
            contig_id = f'{F["Chr_1"]},{F["Pos_1"]},{F["Dir_1"]},{F["SV_ID"]}'
            contig_id2contig[contig_id] = F["Contig"]


    contig_id = None
    with open(output_prefix + ".nanomonsv.rmsk.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = '\t'):

            contig_id = F["Contig_ID"]
            contig_len = int(F["Contig_Len"])

            if contig_id not in contig_id2contig_info: 
                contig_id2contig_info[contig_id] = Contig_info(contig_id, contig_id2contig[contig_id])

            contig_id2contig_info[contig_id].rmsk_info.append(
                Rmsk_record(int(F["Query_Align_Start"]), int(F["Query_Align_End"]), F["Query_Align_Strand"], F["Repeat_NAME"], F["Repeat_Class"], int(F["Repeat_Align_Start"]), int(F["Repeat_Align_End"]))
            )


    with open(output_prefix + ".nanomonsv.bwa.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = '\t'):

            contig_id = F["Contig_ID"]
            contig_len = int(F["Contig_Len"])

            if contig_id not in contig_id2contig_info:
                contig_id2contig_info[contig_id] = Contig_info(contig_id, contig_id2contig[contig_id])

            contig_id2contig_info[contig_id].bwa_info.append(
                Bwa_record(int(F["Query_Align_Start"]), int(F["Query_Align_End"]), F["Query_Align_Strand"], F["Target_Align_Chromosome"], int(F["Target_Align_Start"]), int(F["Target_Align_End"]), int(F["Mapping_Quality"]))
            )

    with open(output_prefix + ".class.txt", 'w') as hin:
        for contig in contig_id2contig_info:
            contig_info = contig_id2contig_info[contig]

            contig_info.check_classification()
            print(f'{contig}\t{contig_info.contig_class}\t{contig_info.sv_key()}', file = hin)


    sv_filterer = Sv_filterer(output_prefix + ".nanomonsv.proc.result.txt", reference_fasta, True, 0.05)
    with open(output_prefix + ".nanomonsv.result.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = "\t"):
            sv_filterer.add_sv(F["Chr_1"], int(F["Pos_1"]), F["Dir_1"], F["Chr_2"], int(F["Pos_2"]), F["Dir_2"],
                F["Inserted_Seq"], F["SV_ID"], int(F["Checked_Read_Num_Tumor"]), int(F["Supporting_Read_Num_Tumor"]),
                int(F["Checked_Read_Num_Control"]), int(F["Supporting_Read_Num_Control"]))

    with open(output_prefix + ".nanomonsv.sbnd.result.txt", 'r') as hin:
        for F in csv.DictReader(hin, delimiter = "\t"):
            contig_id = f'{F["Chr_1"]},{F["Pos_1"]},{F["Dir_1"]},{F["SV_ID"]}'
            if contig_id not in contig_id2contig_info: continue
            contig_info = contig_id2contig_info[contig_id]
            contig_info.check_classification()

            if contig_info.contig_class in ["Plain_SV", "L1_Mediated_Del"]:
                tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = contig_info.sv_key().split(',')
                if tinseq == "---": tinseq = ''
                sv_filterer.add_sv(tchr1, int(tpos1), tdir1, tchr2, int(tpos2), tdir2,
                    tinseq, F["SV_ID"], int(F["Checked_Read_Num_Tumor"]), int(F["Supporting_Read_Num_Tumor"]),
                    int(F["Checked_Read_Num_Control"]), int(F["Supporting_Read_Num_Control"]))
            
    sv_filterer.apply_filters()
    sv_filterer.flush_sv_list()
    del sv_filterer


    with open(output_prefix + ".nanomonsv.sbnd.result.txt", 'r') as hin, open(output_prefix + ".nanomonsv.sbnd.proc.result.txt", 'w') as hout:
        dreader = csv.DictReader(hin, delimiter = '\t')
        print('\t'.join(dreader.fieldnames) + '\t' + "SBND_Class", file = hout)
        for F in dreader:
            contig_id = f'{F["Chr_1"]},{F["Pos_1"]},{F["Dir_1"]},{F["SV_ID"]}'
            if contig_id not in contig_id2contig_info: continue
            contig_info = contig_id2contig_info[contig_id]
            contig_info.check_classification()

            if not contig_info.contig_class in ["Plain_SV", "L1_Mediated_Del"]:
                print('\t'.join(F.values()) + '\t' + contig_info.contig_class, file = hout)
               




#! /usr/bin/env python

import sys

def filter_support_reads(nanomonsv_results, support_read_file, output_file):
    pass_sv_ids = set()
    with open(nanomonsv_results, "r") as f:
        for i, line in enumerate(f):
            if i == 0: continue
            items = line.rstrip("\n").split("\t")
            sv_id = items[7]
            pass_sv_ids.add(sv_id)
    
    with open(support_read_file, "r") as f, open(output_file, "w") as w:
        for line in f:
            items = line.rstrip("\n").split("\t")
            sv_id = items[7]
            if sv_id in pass_sv_ids:
                print(line.rstrip("\n"), file=w)

def get_edge(sorted_support_read_file, output_file, consistent_pos_margin = 1.3):

    temp_rname = ""
    temp_keys = []
    temp_poses = []
    temp_dirs = []
    key2segment_size1 = {}
    key2segment_size2 = {}

    with open(sorted_support_read_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')

            if F[-3] != temp_rname:
                if temp_rname != "" and len(temp_keys) > 1:
                    for i in range(0, len(temp_keys) - 1):
                        # Information of SV_1 and SV_2
                        tchr1_1, tpos1_1, tdir1_1, tchr1_2, tpos1_2, tdir1_2, tinseq1 = temp_keys[i].split(',')
                        tchr2_1, tpos2_1, tdir2_1, tchr2_2, tpos2_2, tdir2_2, tinseq2 = temp_keys[i + 1].split(',')

                        # Read position of SV breakpoint 1
                        jpos1, jpos2 = temp_poses[i], temp_poses[i + 1]
                        # Direction of realignment of reads against SV ref.
                        jdir1, jdir2 = temp_dirs[i], temp_dirs[i + 1]
                        
                        '''
                        Adjust read position of SV breakpoint
                        If realignment direction of SV_1 is "+", then read position is adjusted to read position of breakpoint 2.
                        If realignment direcion of SV_2 is "-", then read position is adjusted to read position of breakpoint 2.
                        '''
                        if jdir1 == '+' and tinseq1 != "---": jpos1 = jpos1 + len(tinseq1)
                        if jdir2 == '-' and tinseq2 != "---": jpos2 = jpos2 - len(tinseq2)
                        
                        # Modify SV order by realignment direction
                        if jdir1 == '+':
                            schr1_1, spos1_1, sdir1_1, schr1_2, spos1_2, sdir1_2 = \
                            tchr1_1, int(tpos1_1), tdir1_1, tchr1_2, int(tpos1_2), tdir1_2
                        else:
                            schr1_2, spos1_2, sdir1_2, schr1_1, spos1_1, sdir1_1 = \
                            tchr1_1, int(tpos1_1), tdir1_1, tchr1_2, int(tpos1_2), tdir1_2

                        if jdir2 == '+':
                            schr2_1, spos2_1, sdir2_1, schr2_2, spos2_2, sdir2_2 = \
                            tchr2_1, int(tpos2_1), tdir2_1, tchr2_2, int(tpos2_2), tdir2_2 
                        else:
                            schr2_2, spos2_2, sdir2_2, schr2_1, spos2_1, sdir2_1 = \
                            tchr2_1, int(tpos2_1), tdir2_1, tchr2_2, int(tpos2_2), tdir2_2

                        sinseq1, sinseq2 = tinseq1, tinseq2

                        # check positional consistency
                        segment_size = '---'
                        consistent_flag = False
                        segment_dir = '---'
                        if schr1_2 == schr2_1:
                            if sdir1_2 == '-' and sdir2_1 == '+': 
                                if spos2_1 - spos1_2 <= consistent_pos_margin * (jpos2 - jpos1) and \
                                spos2_1 - spos1_2 >= (1 / consistent_pos_margin) * (jpos2 - jpos1):
                                    consistent_flag = True
                                    segment_size = spos2_1 - spos1_2
                                    segment_dir = '+'
                            elif sdir1_2 == '+' and sdir2_1 == '-':
                                if spos1_2 - spos2_1 <= consistent_pos_margin * (jpos2 - jpos1) and \
                                spos1_2 - spos2_1 >= (1 / consistent_pos_margin) * (jpos2 - jpos1):
                                    consistent_flag = True
                                    segment_dir = '-'
                                    segment_size = spos1_2 - spos2_1

                        if consistent_flag == True and segment_dir == '-':
                            schr1_1, schr2_2 = schr2_2, schr1_1
                            spos1_1, spos2_2 = spos2_2, spos1_1
                            sdir1_1, sdir2_2 = sdir2_2, sdir1_1
                            schr1_2, schr2_1 = schr2_1, schr1_2
                            spos1_2, spos2_1 = spos2_1, spos1_2
                            sdir1_2, sdir2_1 = sdir2_1, sdir1_2
                            sinseq1, sinseq2 = sinseq2, sinseq1
        
                        if consistent_flag == False: continue
                        #    if schr1_2 < schr2_1 or (schr1_2 == schr2_1 and spos1_2 > spos2_1):
                        #        schr1_1, schr2_2 = schr2_2, schr1_1
                        #        spos1_1, spos2_2 = spos2_2, spos1_1
                        #        sdir1_1, sdir2_2 = sdir2_2, sdir1_1
                        #        schr1_2, schr2_1 = schr2_1, schr1_2
                        #        spos1_2, spos2_1 = spos2_1, spos1_2
                        #        sdir1_2, sdir2_1 = sdir2_1, sdir1_2
                        #        sinseq1, sinseq2 = sinseq2, sinseq1

                        key = '\t'.join([schr1_1, str(spos1_1), sdir1_1, sinseq1, schr1_2, str(spos1_2), sdir1_2,
                                        schr2_1, str(spos2_1), sdir2_1, sinseq2, schr2_2, str(spos2_2), sdir2_2, str(consistent_flag)])

                        if key not in key2segment_size1:
                            key2segment_size1[key], key2segment_size2[key] = str(segment_size), []

                        # key2segment_size1[key].append(str(segment_size))
                        key2segment_size2[key].append(str(jpos2 - jpos1))

                        """
                        print('\t'.join([schr1_1, str(spos1_1), sdir1_1, sinseq1, schr1_2, str(spos1_2), sdir1_2,
                                        schr2_1, str(spos2_1), sdir2_1, sinseq2, schr2_2, str(spos2_2), sdir2_2,
                                        str(consistent_flag), str(segment_size), str(jpos2 - jpos1)]))
                        """

                temp_rname = F[-3]
                temp_keys = []
                temp_poses = []
                temp_dirs = []

            temp_keys.append(",".join(F[0:7]))
            temp_poses.append(int(F[-2]))
            temp_dirs.append(F[-1])

    with open(output_file, "w") as w:
        for key in key2segment_size1:
            segment_size1 = key2segment_size1[key]
            segment_size2 = ','.join(key2segment_size2[key])

            print('\t'.join([key, segment_size1, segment_size2]), file=w)


def connect(edge_file, output_file):
    segments = []
    with open(edge_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            segments.append([F[:16] + ['+']])

    update = True
    while update:
        update = False
        for seg1 in range(len(segments) - 1):
            for seg2 in range(seg1 + 1, len(segments)):
                break1_1 = '\t'.join(segments[seg1][0][:3])
                break1_2 = '\t'.join(segments[seg1][0][4:7])
                break1_3 = '\t'.join(segments[seg1][-1][7:10])
                break1_4 = '\t'.join(segments[seg1][-1][11:14])

                break2_1 = '\t'.join(segments[seg2][0][:3])
                break2_2 = '\t'.join(segments[seg2][0][4:7])
                break2_3 = '\t'.join(segments[seg2][-1][7:10])
                break2_4 = '\t'.join(segments[seg2][-1][11:14])

                if break1_3 == break2_1 and break1_4 == break2_2:
                    sdir1, sdir2 = '+', '-'
                    update = True
                elif break1_3 == break2_4 and break1_4 == break2_3:
                    sdir1, sdir2 = '+', '+'
                    update = True
                elif break1_1 == break2_3 and break1_2 == break2_4:
                    sdir1, sdir2 = '-', '+'
                    update = True
                elif break1_1 == break2_2 and break1_2 == break2_1:
                    sdir1, sdir2 = '-', '-'
                    update = True

                if update: break

            if update: break


        if update:
            if sdir1 == sdir2:
                for j in range(len(segments[seg2])):
                    segments[seg2][j] = segments[seg2][j][11:14] + [segments[seg2][j][10]] + segments[seg2][j][7:10] + \
                                        segments[seg2][j][4:7] + [segments[seg2][j][3]] + segments[seg2][j][:3] + \
                                        segments[seg2][j][14:16] + ['+' if segments[seg2][j][16] == '-' else '-']

                    segments[seg2].reverse()

            if sdir1 == '+':
                segments[seg1] = segments[seg1] + segments[seg2]
            else:
                segments[seg1] = segments[seg2] + segments[seg1]

            del(segments[seg2])
            
    with open(output_file, "w") as w:
        for seg in segments:
            for i, s in enumerate(seg):
                if i == 0:
                    print("\t".join(s), end = "", file=w)
                else:
                    print("\t" + "\t".join(s[7:]), end = "", file=w)
                
                if i == len(seg) - 1:
                    print("", file=w)
    
def main():
    nanomonsv_results = sys.argv[1]
    support_read_file = sys.argv[2]
    filtered_support_read_file = sys.argv[3]
    output_edge_file = sys.argv[4]
    output_connect_file = sys.argv[5]
    
    filter_support_reads(nanomonsv_results, support_read_file, filtered_support_read_file)
    get_edge(filtered_support_read_file, output_edge_file)
    connect(output_edge_file, output_connect_file)

if __name__ == "__main__":
    main()  


"""
8       130933956       +       ---     8       130871365       -       8       130872760       +       ---     8       131401249       +       True    1395 
"""






#!/usr/bin/env python

"""
The developer of nanomonsv borrowed and slightly modify 
the swalign package (https://github.com/mbreese/swalign).

We copy the LINCENSE they provided as below


Copyright (c) 2010-2015 Marcus R. Breese

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.

- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

The copyright holders provide no reassurances that the source code
provided does not infringe any patent, copyright, or any other
intellectual property rights of third parties.  The copyright holders
disclaim any liability to any recipient for claims brought against
recipient by any third party for infringement of that parties
intellectual property rights.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

'''
Simple Smith-Waterman aligner
'''
import sys
from io import StringIO


class ScoringMatrix(object):
    '''
    Read scoring matrix from a file or string

    Matrix should be space-delimited in a format like:

      A C G T
    A 1 0 0 0
    C 0 1 0 0
    G 0 0 1 0
    T 0 0 0 1

    Rows and Columns must be in the same order

    '''
    def __init__(self, filename=None, text=None, wildcard_score=0):
        assert filename or text

        if filename:
            fs = open(filename)
        else:
            fs = StringIO.StringIO(text)

        self.scores = []
        self.bases = None
        self.wildcard_score = wildcard_score

        for line in fs:
            if line[0] == '#' or not line.strip():
                continue

            if not self.bases:
                self.bases = line.split()
                self.base_count = len(self.bases)
            else:
                cols = line.split()
                self.scores.extend([float(x) for x in cols[1:]])

        fs.close()

    def score(self, one, two, wildcard=None):
        if self.wildcard_score and wildcard and (one in wildcard or two in wildcard):
            return self.wildcard_score

        one_idx = 0
        two_idx = 0
        for i, b in enumerate(self.bases):
            if b == one:
                one_idx = i
            if b == two:
                two_idx = i

        return self.scores[(one_idx * self.base_count) + two_idx]


class IdentityScoringMatrix(object):
    def __init__(self, match=1, mismatch=-1):
        self.match = match
        self.mismatch = mismatch

    def score(self, one, two, wildcard=None):
        if wildcard and (one in wildcard or two in wildcard):
            return self.match

        if one == two:
            return self.match
        return self.mismatch

NucleotideScoringMatrix = IdentityScoringMatrix


class Matrix(object):
    def __init__(self, rows, cols, init=None):
        self.rows = rows
        self.cols = cols
        self.values = [init, ] * rows * cols

    def get(self, row, col):
        return self.values[(row * self.cols) + col]

    def set(self, row, col, val):
        self.values[(row * self.cols) + col] = val


class LocalAlignment(object):
    def __init__(self, scoring_matrix, gap_penalty=-1, gap_extension_penalty=-1, gap_extension_decay=0.0, prefer_gap_runs=True, verbose=False, globalalign=False, wildcard=None, full_query=False):
        self.scoring_matrix = scoring_matrix
        self.gap_penalty = gap_penalty
        self.gap_extension_penalty = gap_extension_penalty
        self.gap_extension_decay = gap_extension_decay
        self.verbose = verbose
        self.prefer_gap_runs = prefer_gap_runs
        self.globalalign = globalalign
        self.wildcard = wildcard
        self.full_query = full_query

    def align(self, ref, query, ref_name='', query_name='', rc=False):
        orig_ref = ref
        orig_query = query

        ref = ref.upper()
        query = query.upper()

        matrix = Matrix(len(query) + 1, len(ref) + 1, (0, ' ', 0))
        for row in range(1, matrix.rows):
            matrix.set(row, 0, (0, 'i', 0))

        for col in range(1, matrix.cols):
            matrix.set(0, col, (0, 'd', 0))

        max_val = 0
        max_row = 0
        max_col = 0

        # calculate matrix
        for row in range(1, matrix.rows):
            for col in range(1, matrix.cols):
                mm_val = matrix.get(row - 1, col - 1)[0] + self.scoring_matrix.score(query[row - 1], ref[col - 1], self.wildcard)

                ins_run = 0
                del_run = 0

                if matrix.get(row - 1, col)[1] == 'i':
                    ins_run = matrix.get(row - 1, col)[2]
                    if matrix.get(row - 1, col)[0] == 0:
                        # no penalty to start the alignment
                        ins_val = 0
                    else:
                        if not self.gap_extension_decay:
                            ins_val = matrix.get(row - 1, col)[0] + self.gap_extension_penalty
                        else:
                            ins_val = matrix.get(row - 1, col)[0] + min(0, self.gap_extension_penalty + ins_run * self.gap_extension_decay)
                else:
                    ins_val = matrix.get(row - 1, col)[0] + self.gap_penalty

                if matrix.get(row, col - 1)[1] == 'd':
                    del_run = matrix.get(row, col - 1)[2]
                    if matrix.get(row, col - 1)[0] == 0:
                        # no penalty to start the alignment
                        del_val = 0
                    else:
                        if not self.gap_extension_decay:
                            del_val = matrix.get(row, col - 1)[0] + self.gap_extension_penalty
                        else:
                            del_val = matrix.get(row, col - 1)[0] + min(0, self.gap_extension_penalty + del_run * self.gap_extension_decay)

                else:
                    del_val = matrix.get(row, col - 1)[0] + self.gap_penalty

                if self.globalalign or self.full_query:
                    cell_val = max(mm_val, del_val, ins_val)
                else:
                    cell_val = max(mm_val, del_val, ins_val, 0)

                if not self.prefer_gap_runs:
                    ins_run = 0
                    del_run = 0

                if del_run and cell_val == del_val:
                    val = (cell_val, 'd', del_run + 1)
                elif ins_run and cell_val == ins_val:
                    val = (cell_val, 'i', ins_run + 1)
                elif cell_val == mm_val:
                    val = (cell_val, 'm', 0)
                elif cell_val == del_val:
                    val = (cell_val, 'd', 1)
                elif cell_val == ins_val:
                    val = (cell_val, 'i', 1)
                else:
                    val = (0, 'x', 0)

                # if val[0] >= max_val:
                if val[0] > max_val: # prefer matches with small coordinate (modified by Y.S. on 2022/10/17)
                    max_val = val[0]
                    max_row = row
                    max_col = col

                matrix.set(row, col, val)

        # backtrack
        if self.globalalign:
            # backtrack from last cell
            row = matrix.rows - 1
            col = matrix.cols - 1
            val = matrix.get(row, col)[0]
        elif self.full_query:
            # backtrack from max in last row
            row = matrix.rows - 1
            max_val = 0
            col = 0
            for c in range(1, matrix.cols):
                if matrix.get(row, c)[0] > max_val:
                    col = c
                    max_val = matrix.get(row, c)[0]
            col = matrix.cols - 1
            val = matrix.get(row, col)[0]
        else:
            # backtrack from max
            row = max_row
            col = max_col
            val = max_val

        op = ''
        aln = []

        path = []
        while True:
            val, op, runlen = matrix.get(row, col)

            if self.globalalign:
                if row == 0 and col == 0:
                    break
            elif self.full_query:
                if row == 0:
                    break
            else:
                if val <= 0:
                    break

            path.append((row, col))
            aln.append(op)

            if op == 'm':
                row -= 1
                col -= 1
            elif op == 'i':
                row -= 1
            elif op == 'd':
                col -= 1
            else:
                break

        aln.reverse()

        if self.verbose:
            self.dump_matrix(ref, query, matrix, path)
            print(aln)
            print((max_row, max_col), max_val)

        cigar = _reduce_cigar(aln)
        return Alignment(orig_query, orig_ref, row, col, cigar, max_val, ref_name, query_name, rc, self.globalalign, self.wildcard)

    def dump_matrix(self, ref, query, matrix, path, show_row=-1, show_col=-1):
        sys.stdout.write('      -      ')
        sys.stdout.write('       '.join(ref))
        sys.stdout.write('\n')
        for row in range(matrix.rows):
            if row == 0:
                sys.stdout.write('-')
            else:
                sys.stdout.write(query[row - 1])

            for col in range(matrix.cols):
                if show_row == row and show_col == col:
                    sys.stdout.write('       *')
                else:
                    sys.stdout.write(' %5s%s%s' % (matrix.get(row, col)[0], matrix.get(row, col)[1], '$' if (row, col) in path else ' '))
            sys.stdout.write('\n')


def _reduce_cigar(operations):
    count = 1
    last = None
    ret = []
    for op in operations:
        if last and op == last:
            count += 1
        elif last:
            ret.append((count, last.upper()))
            count = 1
        last = op

    if last:
        ret.append((count, last.upper()))
    return ret


def _cigar_str(cigar):
    out = ''
    for num, op in cigar:
        out += '%s%s' % (num, op)
    return out


class Alignment(object):
    def __init__(self, query, ref, q_pos, r_pos, cigar, score, ref_name='', query_name='', rc=False, globalalign=False, wildcard=None):
        self.query = query
        self.ref = ref
        self.q_pos = q_pos
        self.r_pos = r_pos
        self.cigar = cigar
        self.score = score
        self.r_name = ref_name
        self.q_name = query_name
        self.rc = rc
        self.globalalign = globalalign
        self.wildcard = wildcard

        self.r_offset = 0
        self.r_region = None

        self.orig_query = query
        self.query = query.upper()

        self.orig_ref = ref
        self.ref = ref.upper()

        q_len = 0
        r_len = 0

        self.matches = 0
        self.mismatches = 0

        i = self.r_pos
        j = self.q_pos

        for count, op in self.cigar:
            if op == 'M':
                q_len += count
                r_len += count
                for k in range(count):
                    if self.query[j] == self.ref[i]:
                        self.matches += 1
                    else:
                        self.mismatches += 1
                    i += 1
                    j += 1

            elif op == 'I':
                q_len += count
                j += count
                self.mismatches += count
            elif op == 'D':
                r_len += count
                i += count
                self.mismatches += count

        self.q_end = q_pos + q_len
        self.r_end = r_pos + r_len
        if self.mismatches + self.matches > 0:
            self.identity = float(self.matches) / (self.mismatches + self.matches)
        else:
            self.identity = 0

    def set_ref_offset(self, ref, offset, region):
        self.r_name = ref
        self.r_offset = offset
        self.r_region = region

    @property
    def extended_cigar_str(self):
        qpos = 0
        rpos = 0
        ext_cigar_str = ''
        working = []
        for count, op in self.cigar:
            if op == 'M':
                for k in range(count):
                    if self.query[self.q_pos + qpos + k] == self.ref[self.r_pos + rpos + k]:
                        ext_cigar_str += 'M'
                    else:
                        ext_cigar_str += 'X'
                qpos += count
                rpos += count

            elif op == 'I':
                qpos += count
                ext_cigar_str += 'I' * count
            elif op == 'D':
                rpos += count
                ext_cigar_str += 'D' * count

            working = _reduce_cigar(ext_cigar_str)

        out = ''
        for num, op in working:
            out += '%s%s' % (num, op)
        return out

    @property
    def cigar_str(self):
        return _cigar_str(self.cigar)

    def dump(self, wrap=None, out=sys.stdout):
        i = self.r_pos
        j = self.q_pos

        q = ''
        m = ''
        r = ''
        qlen = 0
        rlen = 0

        for count, op in self.cigar:
            if op == 'M':
                qlen += count
                rlen += count
                for k in range(count):
                    q += self.orig_query[j]
                    r += self.orig_ref[i]
                    if self.query[j] == self.ref[i] or (self.wildcard and (self.query[j] in self.wildcard or self.ref[i] in self.wildcard)):
                        m += '|'
                    else:
                        m += '.'

                    i += 1
                    j += 1
            elif op == 'D':
                rlen += count
                for k in range(count):
                    q += '-'
                    r += self.orig_ref[i]
                    m += ' '
                    i += 1
            elif op == 'I':
                qlen += count
                for k in range(count):
                    q += self.orig_query[j]
                    r += '-'
                    m += ' '
                    j += 1

            elif op == 'N':
                q += '-//-'
                r += '-//-'
                m += '    '

        if self.q_name:
            out.write('Query: %s%s (%s nt)\n' % (self.q_name, ' (reverse-compliment)' if self.rc else '', len(self.query)))
        if self.r_name:
            if self.r_region:
                out.write('Ref  : %s (%s)\n\n' % (self.r_name, self.r_region))
            else:
                out.write('Ref  : %s (%s nt)\n\n' % (self.r_name, len(self.ref)))

        poslens = [self.q_pos + 1, self.q_end + 1, self.r_pos + self.r_offset + 1, self.r_end + self.r_offset + 1]
        maxlen = max([len(str(x)) for x in poslens])

        q_pre = 'Query: %%%ss ' % maxlen
        r_pre = 'Ref  : %%%ss ' % maxlen
        m_pre = ' ' * (8 + maxlen)

        rpos = self.r_pos
        if not self.rc:
            qpos = self.q_pos
        else:
            qpos = self.q_end

        while q and r and m:
            if not self.rc:
                out.write(q_pre % (qpos + 1))  # pos is displayed as 1-based
            else:
                out.write(q_pre % (qpos))  # revcomp is 1-based on the 3' end

            if wrap:
                qfragment = q[:wrap]
                mfragment = m[:wrap]
                rfragment = r[:wrap]

                q = q[wrap:]
                m = m[wrap:]
                r = r[wrap:]
            else:
                qfragment = q
                mfragment = m
                rfragment = r

                q = ''
                m = ''
                r = ''

            out.write(qfragment)
            if not self.rc:
                for base in qfragment:
                    if base != '-':
                        qpos += 1
            else:
                for base in qfragment:
                    if base != '-':
                        qpos -= 1

            if not self.rc:
                out.write(' %s\n' % qpos)
            else:
                out.write(' %s\n' % (qpos + 1))

            out.write(m_pre)
            out.write(mfragment)
            out.write('\n')
            out.write(r_pre % (rpos + self.r_offset + 1))
            out.write(rfragment)
            for base in rfragment:
                if base != '-':
                    rpos += 1
            out.write(' %s\n\n' % (rpos + self.r_offset))

        out.write("Score: %s\n" % self.score)
        out.write("Matches: %s (%.1f%%)\n" % (self.matches, self.identity * 100))
        out.write("Mismatches: %s\n" % (self.mismatches,))
        out.write("CIGAR: %s\n" % self.cigar_str)


def fasta_gen(fname):
    def gen():
        seq = ''
        name = ''
        comments = ''

        if fname == '-':
            f = sys.stdin
            name = 'stdin'
        else:
            f = open(fname)

        for line in f:
            if line[0] == '>':
                if name and seq:
                    yield (name, seq, comments)

                spl = line[1:].strip().split(' ', 1)
                name = spl[0]
                if len(spl) > 1:
                    comments = spl[1]
                else:
                    comments = ''

                seq = ''
            else:
                seq += line.strip()

        if name and seq:
            yield (name, seq, comments)

        if fname != '-':
            f.close()
    return gen


def seq_gen(name, seq):
    def gen():
        yield (name, seq, '')

    return gen


def extract_region(comments):
    ref = None
    start = None
    # start_offset = 0
    # end_offset = 0

    try:
        attrs = comments.split(' ')
        for attr in attrs:
            if '=' in attr:
                k, v = attr.split('=')
                if k == 'range':
                    spl = v.split(':')
                    ref = spl[0]
                    start, end = [int(x) for x in spl[1].split('-')]
                # elif k == "5'pad":
                #     start_offset = int(v)
                # elif k == "3'pad":
                #     end_offset = int(v)
    except:
        pass

    if ref and start:
        return (ref, start - 1, '%s:%s-%s' % (ref, start, end))

    return None


__revcomp = {}
for a, b in zip('atcgATCGNn', 'tagcTAGCNn'):
    __revcomp[a] = b
__cache = {}


def revcomp(seq):
    if seq in __cache:
        return __cache[seq]

    ret = []
    for s in seq.upper()[::-1]:
        ret.append(__revcomp[s])

    __cache[seq] = ''.join(ret)
    return __cache[seq]


#     sw.align('ACACACTA','AGCACACA').dump()
#     aln=sw.align("AAGGGGAGGACGATGCGGATGTTC","AGGGAGGACGATGCGG")
#     aln.dump()



#! /usr/bin/env python3

import sys, os, subprocess, itertools
import pysam

from .logger import get_logger
from .utils import get_alignment_object 
logger = get_logger(__name__)

def parse_alignment_info(input_alignment_file, reference_fasta, deletion_output_file, insertion_output_file, rearrangement_output_file,
    breakpoint_output_file, min_ins_size = 20, min_del_size = 30, min_clipping_size_for_bp = 200):

    """Parse BAM file to obtain putative SV supporting reads and their associated info."""

    hout_d = open(deletion_output_file, 'w')
    hout_i = open(insertion_output_file, 'w')
    hout_r = open(rearrangement_output_file, 'w') 
    hout_b = open(breakpoint_output_file, 'w')

    alignment_h = get_alignment_object(input_alignment_file, reference_fasta)
    
    for read in alignment_h.fetch():

        # if read.is_secondary: continue

        query_name = read.query_name
        query_strand = '-' if read.is_reverse else '+'
        query_length = read.infer_read_length()

        reference_name = read.reference_name
        reference_start = read.reference_start + 1
        reference_end = read.reference_end
        mapping_quality = str(read.mapping_quality)
        is_secondary = read.is_secondary
        is_supplementary = read.is_supplementary

        # import pdb; pdb.set_trace()
        cigar_stats = read.get_cigar_stats()
        num_M = cigar_stats[0][0] + cigar_stats[0][7] + cigar_stats[0][8]
        num_I = cigar_stats[0][1]
        num_D = cigar_stats[0][2]

        cigartuples = read.cigartuples
        left_hard_clipping_size, right_hard_clipping_size = 0, 0
        left_soft_clipping_size, right_soft_clipping_size = 0, 0
        if cigartuples[0][0] == 5: left_hard_clipping_size = cigartuples[0][1]
        if cigartuples[-1][0] == 5: right_hard_clipping_size = cigartuples[-1][1]
        if cigartuples[0][0] == 4: left_soft_clipping_size = cigartuples[0][1]
        if cigartuples[-1][0] == 4: right_soft_clipping_size = cigartuples[-1][1]

        if not is_supplementary:
            if query_strand == '+':
                query_start = read.query_alignment_start + 1 
                query_end = read.query_alignment_end
            else:
                query_start = query_length - read.query_alignment_end + 1
                query_end = query_length - read.query_alignment_start
        else:
            if query_strand == '+':
                query_start = left_hard_clipping_size + left_soft_clipping_size + 1
                query_end = query_length - right_hard_clipping_size - right_soft_clipping_size
            else:
                query_start = right_hard_clipping_size + right_soft_clipping_size + 1
                query_end = query_length - left_hard_clipping_size - left_soft_clipping_size 


        query_pos_cur = query_start - 1 if query_strand == '+' else query_end
        query_pos_check = query_pos_cur
        reference_pos_cur = reference_start - 1
        reference_pos_check = reference_start - 1

        for cigar in cigartuples:
            if cigar[0] in [0, 7, 8]:
                query_pos_cur = query_pos_cur + cigar[1] if query_strand == '+' else query_pos_cur - cigar[1]
                reference_pos_cur = reference_pos_cur + cigar[1] 
            elif cigar[0] == 1:

                if cigar[1] >= min_ins_size:

                    tinfo = f'{query_start},{query_pos_cur},{query_end},{query_length},{query_strand},' + \
                        f'{mapping_quality},{num_M},{num_I - cigar[1]},{num_D},{is_supplementary},{is_secondary}'

                    print(f'{reference_name}\t{reference_pos_cur}\t{reference_pos_cur + 1}\t{query_name}\t{cigar[1]}\t+\t{tinfo}',
                        file = hout_i)

                query_pos_cur = query_pos_cur + cigar[1] if query_strand == '+' else query_pos_cur - cigar[1]

            elif cigar[0] == 2: 
                if cigar[1] >= min_del_size:

                    """
                    if query_strand == '+':
                        query_start2 = query_pos_check + 1
                        query_end2 = query_pos_cur
                        reference_start2 = reference_pos_check + 1
                        reference_end2 = reference_pos_cur
                    else:   
                        query_start2 = query_pos_cur + 1
                        query_end2 = query_pos_check
                        reference_start2 = reference_pos_check + 1
                        reference_end2 = reference_pos_cur
                    """

                    tinfo = f'{query_start},{query_pos_cur},{query_end},{query_length},{query_strand},' + \
                        f'{mapping_quality},{num_M},{num_I},{num_D - cigar[1]},{is_supplementary},{is_secondary}'

                    print(f'{reference_name}\t{reference_pos_cur}\t{reference_pos_cur + cigar[1]}\t{query_name}\t{cigar[1]}\t+\t{tinfo}',
                        file = hout_d)

                    reference_pos_cur = reference_pos_cur + cigar[1]
                else:
                    reference_pos_cur = reference_pos_cur + cigar[1]

        if query_strand == '+' and query_end != query_pos_cur:
            logger.error("query end inconsistent!! %s: %d != %d" % (query_name, query_end, query_pos_cur))
            sys.exit(1)
        if query_strand == '-' and query_start != query_pos_cur + 1:
            logger.error("query end inconsistent!! %s: %d != %d" % (query_name, query_end, query_pos_cur))
            sys.exit(1)

        """
        if query_strand == '+':
            query_start2 = query_pos_check + 1
            query_end2 = query_pos_cur
            reference_start2 = reference_pos_check + 1
            reference_end2 = reference_pos_cur
        else:   
            query_start2 = query_pos_cur + 1
            query_end2 = query_pos_check
            reference_start2 = reference_pos_check + 1
            reference_end2 = reference_pos_cur

        import pdb; pdb.set_trace() 
        if query_start != query_start2:
            print("query start inconsistent!!")
            sys.exit(1)
        if query_end != query_end2:
            print("query end inconsistent!!")
            sys.exit(1)
        if reference_start != reference_start2:
            print("reference start inconsistent!!")
            sys.exit(1)
        if reference_end != reference_end2:
            print("reference end inconsistent!!")
            sys.exit(1)
        """

        print(f'{query_name}\t{query_start}\t{query_end}\t{query_length}\t{query_strand}\t{reference_name}\t' + \
            f'{reference_start}\t{reference_end}\t{mapping_quality}\t{num_M}\t{num_I}\t{num_D}\t' + \
            f'{is_supplementary}\t{is_secondary}', file = hout_r)

        # print('\t'.join([query_name, str(query_start), str(query_end), str(query_length), query_strand, \
        #                  reference_name, str(reference_start), str(reference_end), mapping_quality, \
        #                  str(num_M), str(num_I), str(num_D), str(is_supplementary), str(is_secondary)]), file = hout_r)


        if left_soft_clipping_size + left_hard_clipping_size >= min_clipping_size_for_bp:

            if query_strand == '+':
                tinfo = f'{query_start},{query_start},{query_end},{query_length},{query_strand},' + \
                    f'{mapping_quality},{num_M},{num_I},{num_D},{is_supplementary},{is_secondary}'
            else:
                tinfo = f'{query_start},{query_end},{query_end},{query_length},{query_strand},' + \
                    f'{mapping_quality},{num_M},{num_I},{num_D},{is_supplementary},{is_secondary}'
 
            print(f'{reference_name}\t{reference_start - 1}\t{reference_start}\t{query_name}\t*\t-\t{tinfo}',
                file = hout_b)

        if right_soft_clipping_size + right_hard_clipping_size >= min_clipping_size_for_bp:
            
            if query_strand == '+':
                tinfo = f'{query_start},{query_end},{query_end},{query_length},{query_strand},' + \
                    f'{mapping_quality},{num_M},{num_I},{num_D},{is_supplementary},{is_secondary}'
            else:
                tinfo = f'{query_start},{query_start},{query_end},{query_length},{query_strand},' + \
                    f'{mapping_quality},{num_M},{num_I},{num_D},{is_supplementary},{is_secondary}'

            print(f'{reference_name}\t{reference_end - 1}\t{reference_end}\t{query_name}\t*\t+\t{tinfo}',
                file = hout_b)


    hout_d.close()
    hout_i.close()
    hout_r.close()
    hout_b.close()
    alignment_h.close()


def extract_bedpe_junction(input_file, output_file, split_alignment_check_margin1 = 50, split_alignment_check_margin2 = 50, minimum_ambiguity = 20):

   
    def print_bedpe_junction(query2target, hout):

        query_list = list(query2target)
        for qpos_comb in list(itertools.combinations(query_list, 2)):

            if qpos_comb[0][1] < qpos_comb[1][1]:
                qpos1, qpos2 = qpos_comb[0], qpos_comb[1]
            else:
                qpos1, qpos2 = qpos_comb[1], qpos_comb[0]
            
            # if the first region completely covers the second region
            if qpos1[2] >= qpos2[2]: continue
            
            # if there is a significant overlap         
            if (qpos1[2] - qpos2[1]) / (qpos2[2] - qpos1[2]) >= 0.2: continue
            
            # if abs(qpos2[1] - qpos1[2]) <= split_alignment_check_margin:
            if qpos2[1] - qpos1[2] <= split_alignment_check_margin1 and qpos1[2] - qpos2[1] <= split_alignment_check_margin2:
                bp_flag = True
                tchr1, tstart1, tend1, tmapQ1, tnumM1, tnumI1, tnumD1, tis_supp1, tis_2nd1 = query2target[qpos1]
                tchr2, tstart2, tend2, tmapQ2, tnumM2, tnumI2, tnumD2, tis_supp2, tis_2nd2 = query2target[qpos2]
                
                if qpos2[1] - qpos1[2] > 0:
                    outward_ambiguity, inward_ambiguity = max(qpos2[1] - qpos1[2], minimum_ambiguity), minimum_ambiguity
                else:
                    outward_ambiguity, inward_ambiguity = minimum_ambiguity, max(qpos1[2] - qpos2[1], minimum_ambiguity)
                
                
                bchr1, bchr2 = tchr1, tchr2
                if qpos1[4] == '+': 
                    bstart1, bend1, bstrand1 = max(int(tend1) - inward_ambiguity, 0), int(tend1) + outward_ambiguity, '+'
                else:
                    bstart1, bend1, bstrand1 = max(int(tstart1) - outward_ambiguity, 0), int(tstart1) + inward_ambiguity, '-'
                
                if qpos2[4] == '+': 
                    bstart2, bend2, bstrand2 = max(int(tstart2) - outward_ambiguity, 0), int(tstart2) + inward_ambiguity, '-'
                else:
                    bstart2, bend2, bstrand2 = max(int(tend2) - inward_ambiguity, 0), int(tend2) + outward_ambiguity, '+'

                
                bread_name = qpos1[0]
                
                binfo1 = ','.join([str(qpos1[1]), '*', str(qpos1[2]), str(qpos1[3]), qpos1[4], tmapQ1, tnumM1, tnumI1, tnumD1, tis_supp1, tis_2nd1])
                binfo2 = ','.join([str(qpos2[1]), '*', str(qpos2[2]), str(qpos2[3]), qpos2[4], tmapQ2, tnumM2, tnumI2, tnumD2, tis_supp2, tis_2nd2])
                
                if bchr1 > bchr2 or (bchr1 == bchr2 and bstart1 > bstart2):
                    bchr1, bstart1, bend1, bstrand1, binfo1, bchr2, bstart2, bend2, bstrand2, binfo2 = \
                        bchr2, bstart2, bend2, bstrand2, binfo2, bchr1, bstart1, bend1, bstrand1, binfo1
               
                print(f'{bchr1}\t{bstart1}\t{bend1}\t{bchr2}\t{bstart2}\t{bend2}\t{bread_name}\t0\t{bstrand1}\t{bstrand2}\t{binfo1}\t{binfo2}',
                    file = hout)

 
    hout = open(output_file, 'w')

    temp_read_name = ''
    query2target = {}
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            # skip secondary alignment
            # if F[13] == "True": continue

            if F[0] != temp_read_name:
                if temp_read_name != '' and len(query2target) > 1: print_bedpe_junction(query2target, hout)
                            
                temp_read_name = F[0]
                query2target = {}

            query2target[(F[0], int(F[1]), int(F[2]), int(F[3]), F[4])] = (F[5], int(F[6]), int(F[7]), F[8], F[9], F[10], F[11], F[12], F[13])

    if temp_read_name != '' and len(query2target) > 1: print_bedpe_junction(query2target, hout)

    hout.close()

 
if __name__ == "__main__":

    import sys

    input_bam = sys.argv[1]
    output_prefix = sys.argv[2]

    parse_alignment_info(input_bam, output_prefix + ".tmp.alignment_info.txt")

    hout = open(output_prefix + ".tmp.alignment_info.name_sorted.txt", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", output_prefix + ".tmp.alignment_info.txt"], stdout = hout)
    hout.close()

    extract_bedpe_junction(output_prefix + ".tmp.alignment_info.name_sorted.txt", output_prefix + ".tmp.junction.bedpe")

    hout = open(output_prefix + ".tmp.junction.sorted.bedpe", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k4,4", "-k5,5n", "-k6,6n", output_prefix + ".tmp.junction.bedpe"], stdout = hout)
    hout.close()
  
    hout = open(output_prefix + ".junction.sorted.bedpe.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", output_prefix + ".tmp.junction.sorted.bedpe"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", output_prefix + ".junction.sorted.bedpe.gz"])







#! /usr/bin/env python3

import sys, os, subprocess, shutil, statistics 
import pysam

from . import smith_waterman
from .my_seq import reverse_complement
from .logger import get_logger

logger = get_logger(__name__)


def get_refined_bp(contig, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, h_log, 
                   sw_jump_params, rd_margin = 20, i_margin = 500):

    start1 = max(1, start1)
    start2 = max(1, start2)

    if mode != "i":

        ref_len1 = fasta_file_ins.get_reference_length(chr1)
        ref_len2 = fasta_file_ins.get_reference_length(chr2)

        bstart1, bend1 = max(int(start1) - rd_margin, 1), int(end1) + rd_margin
        bstart2, bend2 = max(int(start2) - rd_margin, 1), int(end2) + rd_margin

        if ref_len1 < bend1: bend1 = ref_len1
        if ref_len2 < bend2: bend2 = ref_len2    

        region1_seq = fasta_file_ins.fetch(chr1, bstart1 - 1, bend1).upper()
        region2_seq = fasta_file_ins.fetch(chr2, bstart2 - 1, bend2).upper()
        

        if dir1 == '-': region1_seq = reverse_complement(region1_seq)
        if dir2 == '+': region2_seq = reverse_complement(region2_seq)

        sret = smith_waterman.sw_jump(contig, region1_seq, region2_seq,
                                      match_score = sw_jump_params[0], mismatch_penalty = sw_jump_params[1],
                                      gap_cost = sw_jump_params[2], jump_cost = sw_jump_params[3])

        if sret is None: return(None)
        score, contig_align, region1_align, region2_align, contig_seq, region_seq = sret

        bp_pos1 = bstart1 + region1_align[1] - 1 if dir1 == '+' else bend1 - region1_align[1] + 1 
        bp_pos2 = bstart2 + region2_align[0] - 1 if dir2 == '-' else bend2 - region2_align[0] + 1

        if contig_align[2] - contig_align[1] == 1:
            inseq = '---'
        elif contig_align[2] - contig_align[1] > 1:
            inseq = contig[(contig_align[1]):(contig_align[2] - 1)]
            if dir1 == '-': inseq = reverse_complement(inseq)
        else:
            logger.warning("Alignment inconsistent!!")

        print(score, contig_align, region1_align, region2_align, file = h_log)
        print(contig_seq, file = h_log)
        print(region_seq, file = h_log)

        return(bp_pos1, bp_pos2, inseq)
    
    else:
    
        contig_start = contig[:min(i_margin, len(contig))]
        contig_end = contig[-min(i_margin, len(contig)):]

        region_seq = fasta_file_ins.fetch(chr1, max(0, start1 - 100), end2 + 100).upper()
        sret = smith_waterman.sw_jump(region_seq, contig_start, contig_end,
                                      match_score = sw_jump_params[0], mismatch_penalty = sw_jump_params[1], 
                                      gap_cost = sw_jump_params[2], jump_cost = sw_jump_params[3])

        if sret is None: return(None)
        score, region_align, contig_start_align, contig_end_align, region_seq, contig_seq = sret

        bp_pos1 = max(0, start1 - 100) + region_align[1]
        bp_pos2 = max(0, start1 - 100) + region_align[2]

        inseq_start = contig_start_align[1]
        inseq_end = len(contig) - (len(contig_end) - contig_end_align[0] + 1)
        inseq = contig[inseq_start:(inseq_end + 1)]

        print(score, region_align, contig_start_align, contig_end_align, file = h_log)
        print(region_seq, file = h_log)
        print(contig_seq, file = h_log)

        return(bp_pos1, bp_pos2, inseq)

        
 
def locate_bp(consensus_file, output_file, reference_fasta, sw_jump_params, debug):

    fasta_file_ins = pysam.FastaFile(reference_fasta)

    with open(consensus_file, 'r') as hin, open(output_file, 'w') as hout, \
        open(output_file + ".locate_bp.log", 'w') as hout_log:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            temp_key, tconsensus = F[0], F[1]
            print(temp_key + '\n' + tconsensus, file = hout_log)
            chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, cid = temp_key.split(',')
            start1, end1, start2, end2 = int(start1), int(end1), int(start2), int(end2)       
            bret = get_refined_bp(tconsensus, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, hout_log, sw_jump_params)
            if bret is not None:  
                bp_pos1, bp_pos2, inseq = bret 
                print(bp_pos1, bp_pos2, inseq, file = hout_log)
                print('', file = hout_log)
                print(f"{chr1}\t{bp_pos1}\t{dir1}\t{chr2}\t{bp_pos2}\t{dir2}\t{inseq}\t{mode}_{cid}", file = hout)
                # print('\t'.join([chr1, str(bp_pos1), dir1, chr2, str(bp_pos2), dir2, inseq]), file = hout)

    if not debug: os.remove(output_file + ".locate_bp.log")
